/**************************************************************************
The program accepts a circuit description in ISCAS'85 format and outputs 
the equivalent CNF formula (in DIMACS format). 
***************************************************************************/

int main(int argc,char *argv[])
{
  char *NameOfFile;
  int clause_count = 0;
  int var_count = 1;
  int out_idx = 0;
  FILE *tmpFile;
  std::ifstream* myInputFileStream;
  tmpFile = fopen("iscas2cnf.tmp", "rw");
  if(!tmpFile)
  {
    printf("c Cannot open tmp file : iscas2cnf.tmp\n");
    exit(-1);
  }

  if(argc != 2)
  {
    printf("c ISCAS'85 to CNF file converter\n");
    printf("Usage: iscas2cnf input-file\n");
    exit (-2);
  }
  NameOfFile = argv[1];
  myInputFileStream= new std::ifstream(NameOfFile);
  
  if (!(*myInputFileStream))
  {
    printf("c Cannot open tmp file : %s\n",NameOfFile);
    exit(-3);
  }
 
  boost::char_separator<char> sep("\t ", "=(,)");
  std::string str;
  while (getline(*myInputFileStream, str))
  {
    myStringNumber++;

    if (str.size() == 0)
    {
      printf("c Error: empty string in input file\n");
      exit(-4);
    }
    
    if (str[0] == '#')
      continue;

    boost::tokenizer<boost::char_separator<char> > tok(str,sep); // all tokens of the cuurent string
    boost::tokenizer<boost::char_separator<char> >::iterator cur_tok = tok.begin(); // current token
    
    if (*cur_tok == "INPUT" || *cur_tok == "input")
    {
      cur_tok++; if (cur_tok == tok.end()) this->reportFailure("unexpected end of string");
    }
    if (/INPUT\s*\(\s*([\w\._]+)\s*\)/) {
	if ($opt_d) { print "INPUT = $1\n"; }
	if ($opt_d) { print "DEFINING $var_count FOR $1!\n"; }
	$node{$1} = $var_count++;       # Specify variable
    }
    elsif (/input\s*\(\s*([\w\._]+)\s*\)/) {
	if ($opt_d) { print "INPUT = $1\n"; }
	if ($opt_d) { print "DEFINING $var_count FOR $1!\n"; }
	$node{$1} = $var_count++;       # Specify variable
    }
    elsif (/OUTPUT\s*\(\s*([\w\._]+)\s*\)/) {
	if ($opt_d) { print "OUTPUT = $1\n"; }
	if ($opt_d) { print "DEFINING $var_count FOR $1!\n"; }
	$out_vars[$out_idx++] = $var_count;
	$node{$1} = $var_count++;
    }
    elsif (/output\s*\(\s*([\w\._]+)\s*\)/) {
	if ($opt_d) { print "OUTPUT = $1\n"; }
	if ($opt_d) { print "DEFINING $var_count FOR $1!\n"; }
	$out_vars[$out_idx++] = $var_count;
	$node{$1} = $var_count++;
    }

    if (/([\w\._]+)\s*=\s*(\w+)\s*\(\s*(.+)\s*\)/) {
	if (!defined($node{$1})) {
	    if ($opt_d) { print "DEFINING $var_count FOR $1!\n"; }
	    $node{$1} = $var_count++;
	}

	$out_node = $1;
	$g_type = $2;
	$in_node_str = $3;

	@inp_def = split(/\s*,\s*/, $in_node_str);

	@tmp_def = split(/\s+/, $inp_def[$#inp_def]); # Remove trailing spaces
	$inp_def[$#inp_def] = $tmp_def[0];

	for ($i=0;$i<=$#inp_def;$i++) {
	    if (!defined($node{$inp_def[$i]})) {
		$node{$inp_def[$i]} = $var_count++;
		###die "Gate $out_node uses input $inp_def[$i] which is not yet defined. Exiting...\n";
	    }
	}

	if ($g_type eq "AND" || $g_type eq "and" ||
	    $g_type eq "NAND" || $g_type eq "nand" ||
	    $g_type eq "OR" || $g_type eq "or" ||
	    $g_type eq "NOR" || $g_type eq "nor") {
	    $clause_count += $#inp_def + 2;
	}
	elsif ($g_type eq "NOT" || $g_type eq "not" ||
	       $g_type eq "BUFF" || $g_type eq "buff") {
	    $clause_count += 2;
	}
	elsif ($g_type eq "XOR" || $g_type eq "xor" ||
	       $g_type eq "XNOR" || $g_type eq "xnor") {
	    $clause_count += 4;
	}

	if ($opt_d) {
	    print "IN STR = $in_node_str\n";
	    print "OUT = $out_node\n";
	    print "GATE = $g_type\n";
	    print "\t\t(@inp_def) W/ $#inp_def inputs\n";
	}

	if ($g_type eq "AND" || $g_type eq "and") {
	    &write_AND_gate (@inp_def, $out_node);
	}
	elsif ($g_type eq "NAND" || $g_type eq "nand") {
	    &write_NAND_gate (@inp_def, $out_node);
	}
  	elsif ($g_type eq "OR" || $g_type eq "or") {
  	    &write_OR_gate (@inp_def, $out_node);
  	}
  	elsif ($g_type eq "NOR" || $g_type eq "nor") {
  	    &write_NOR_gate (@inp_def, $out_node);
  	}
  	elsif ($g_type eq "NOT" || $g_type eq "not") {
  	    &write_NOT_gate (@inp_def, $out_node);
  	}
  	elsif ($g_type eq "BUFF" || $g_type eq "buff") {
  	    &write_BUFF_gate (@inp_def, $out_node);
  	}
  	elsif ($g_type eq "XOR" || $g_type eq "xor") {
  	    &write_XOR_gate (@inp_def, $out_node);
  	}
  	elsif ($g_type eq "XNOR" || $g_type eq "xnor") {
  	    &write_XNOR_gate (@inp_def, $out_node);
 	}
    }
}

if (defined($opt_o)) { # Count objectives defined in file -> update cl number
    open(OBJF, "<$opt_o") || die "Could not open output file $opt_o\n";
    while (<OBJF>) {
	if (/(\w+)\s+([01])/) { $clause_count++; }
    }
    close (OBJF);
}
$var_count--;

print "c Output from iscas2cnf\n";
if (defined($opt_O)) { # Update number of clauses
    $clause_count++;
}
print "p cnf $var_count $clause_count\n";

close TMPF;
open(TMPF, "<$tmp_fname") || die "Could not open output file $tmp_fname\n";

while(<TMPF>) {
    print;
}
system("rm $tmp_fname");

if (defined($opt_o)) { # Objectives defined in file
    open(OBJF, "<$opt_o") || die "Could not open output file $opt_o\n";
    while (<OBJF>) {
	if (/(\w+)\s+([01])/) {
	    if ($opt_d) { print "$1\n"; }
	    if (!$2) {
		print "-";
	    }
	    print "$node{$1} 0\n";
	}
    }
    close (OBJF);
}
elsif (defined($opt_O)) {
    for ($i=0; $i<$out_idx; $i++) {
	if ($opt_O == 0) {
	    print "-";
	}
	print "$out_vars[$i] 0\n";
    }
}

1;


sub write_AND_gate {
    local (@nodes) = @_;
    local $i;
    local $out;

    $out = $nodes[$#nodes];
    if ($opt_d) { print "PRINTING $out = AND @nodes!\n"; }

    for ($i = 0; $i<$#nodes; $i++) {
	print TMPF "$node{$nodes[$i]} -$node{$out} 0\n";
    }
    print TMPF "$node{$out} ";
    for ($i = 0; $i<$#nodes; $i++) {
	print TMPF "-$node{$nodes[$i]} ";
    }
    print TMPF "0\n";
}

sub write_NAND_gate {
    local (@nodes) = @_;
    local $i;
    local $out;

    $out = $nodes[$#nodes];
    if ($opt_d) { print "PRINTING $out = NAND @nodes!\n"; }

    for ($i = 0; $i<$#nodes; $i++) {
	print TMPF "$node{$nodes[$i]} $node{$out} 0\n";
    }
    print TMPF "-$node{$out} ";
    for ($i = 0; $i<$#nodes; $i++) {
	print TMPF "-$node{$nodes[$i]} ";
    }
    print TMPF "0\n";
}


sub write_OR_gate {
    local (@nodes) = @_;
    local $i;
    local $out;

    $out = $nodes[$#nodes];
    if ($opt_d) { print "PRINTING $out = OR @nodes!\n"; }

    for ($i = 0; $i<$#nodes; $i++) {
	print TMPF "-$node{$nodes[$i]} $node{$out} 0\n";
    }
    print TMPF "-$node{$out} ";
    for ($i = 0; $i<$#nodes; $i++) {
	print TMPF "$node{$nodes[$i]} ";
    }
    print TMPF "0\n";
}

sub write_NOR_gate {
    local (@nodes) = @_;
    local $i;
    local $out;

    $out = $nodes[$#nodes];
    if ($opt_d) { print "PRINTING $out = NOR @nodes!\n"; }

    for ($i = 0; $i<$#nodes; $i++) {
	print TMPF "-$node{$nodes[$i]} -$node{$out} 0\n";
    }
    print TMPF "$node{$out} ";
    for ($i = 0; $i<$#nodes; $i++) {
	print TMPF "$node{$nodes[$i]} ";
    }
    print TMPF "0\n";
}


sub write_NOT_gate {
    local (@nodes) = @_;
    local $out;

    $out = $nodes[$#nodes];
    if ($opt_d) { print "PRINTING $out = NOT @nodes!\n"; }

    print TMPF "-$node{$nodes[0]} -$node{$out} 0\n";
    print TMPF "$node{$nodes[0]} $node{$out} 0\n";
}

sub write_BUFF_gate {
    local (@nodes) = @_;
    local $out;

    $out = $nodes[$#nodes];
    if ($opt_d) { print "PRINTING $out = BUFF @nodes!\n"; }

    print TMPF "$node{$nodes[0]} -$node{$out} 0\n";
    print TMPF "-$node{$nodes[0]} $node{$out} 0\n";
}


sub write_XOR_gate {
    local (@nodes) = @_;
    local $out;

    $out = $nodes[$#nodes];
    if ($opt_d) { print "PRINTING $out = NOT @nodes!\n"; }

    print TMPF "-$node{$nodes[0]} $node{$nodes[1]} $node{$out} 0\n";
    print TMPF "$node{$nodes[0]} -$node{$nodes[1]} $node{$out} 0\n";

    print TMPF "-$node{$nodes[0]} -$node{$nodes[1]} -$node{$out} 0\n";
    print TMPF "$node{$nodes[0]} $node{$nodes[1]} -$node{$out} 0\n";
}

sub write_XNOR_gate {
    local (@nodes) = @_;
    local $out;

    $out = $nodes[$#nodes];
    if ($opt_d) { print "PRINTING $out = NOT @nodes!\n"; }

    print TMPF "-$node{$nodes[0]} $node{$nodes[1]} -$node{$out} 0\n";
    print TMPF "$node{$nodes[0]} -$node{$nodes[1]} -$node{$out} 0\n";

    print TMPF "-$node{$nodes[0]} -$node{$nodes[1]} $node{$out} 0\n";
    print TMPF "$node{$nodes[0]} $node{$nodes[1]} $node{$out} 0\n";
}


